/* Gruppe 11
 *
 * Autoren: Daniel Pujiula
 */
package grp11;

import trigrams.*;

/**
 * TrigramStorage that uses a combination of HashMap, PriorityHeap and pre-calculated sums to answer
 * Queries as efficient as possible.
 */
public class AVLPriorityHeapTrigramStorage implements TrigramStorage {

    static final int HASHTABLELENGTH = 100000000;
//    static final boolean DEBUG = true;

    AVLPriorityHeap bst = new AVLPriorityHeap();
    AVLPriorityHeap[] hashMap = new AVLPriorityHeap[HASHTABLELENGTH];
    int count = 0;

    private int computeHash(Trigram trigram) {
        return  Math.abs((trigram.getW1().hashCode() + 7*trigram.getW2().hashCode() + 15*trigram.getW3().hashCode())  ) % HASHTABLELENGTH;
//        System.out.println(hash);
    }

    @Override
    public void insert(TrigramAndCount trigram) {
        // Insert into priority heap
        AVLPriorityHeapNode node = new AVLPriorityHeapNode(trigram.getW1(),
                trigram.getW2(), trigram.getW3(), trigram.getCount());
        bst.priorityHeapInsert(node);
        // Insert into hash table
        int hash = computeHash(trigram);
        if(hashMap[hash] == null) hashMap[hash] = new AVLPriorityHeap();
        hashMap[hash].hashInsert(node);
    }

    @Override
    public void build() {
        // accumulate all counts
        AVLPriorityHeapNode min = bst.bstMin(bst.root);
        long cumCount = 0;
        while(min != null) {
            ++count;
            cumCount += min.getCount();
            min.cumCount = cumCount;
            min.id = count;
            min = bst.bstSucc(min);
        }
/*
        if(DEBUG) {
            // Build histogram
            int maxheight = 0;
            for (int i = 0; i < HASHTABLELENGTH; i++) {
                if (hashMap[i] != null) maxheight = Math.max(maxheight, hashMap[i].hash_height());
            }
            int[] histogram = new int[maxheight + 1];
            for (int i = 0; i < HASHTABLELENGTH; i++) {
                int h = 0;
                if (hashMap[i] != null) h = hashMap[i].hash_height();
                histogram[h]++;
            }
            for (int i = 0; i <= maxheight; i++) {
                System.out.println("H " + i + " -> " + histogram[i]);
            }
        }
*/
    }

    private AVLPriorityHeapNode getNearestNeighbour(Trigram trigram, boolean before) {
        AVLPriorityHeapNode node = null;
        int hash = computeHash(trigram);
        if(hashMap[hash] != null) node = hashMap[hash].hashSearch(trigram);
        if(node == null) node = bst.bstSearchNearestNeighbour(trigram, before);
        return node;
    }

    @Override
    public QueryResult query(Trigram trigram1, Trigram trigram2) {
        long max = 0;
        long sum = 0;

        // long start = System.nanoTime();

        AVLPriorityHeapNode node1 = getNearestNeighbour(trigram1, false);
        AVLPriorityHeapNode node2 = getNearestNeighbour(trigram2, true);

        //    int hops = 0;
        //    long middle = System.nanoTime();

        if (/*trigram1 == null || trigram2 == null || */ node1 == null || node2 == null || node1.id > node2.id) {
            return new QueryResult(max, sum);
        }

        sum = node2.cumCount - node1.cumCount + node1.getCount();

        AVLPriorityHeapNode traverser = node1;

        max = traverser.getCount();
        traverser = traverser.parent;

        // find maximum
        while(traverser != null && traverser.id <= node2.id) {
            //            ++hops;
            if(traverser.id > node1.id) max = traverser.getCount();
            traverser = traverser.parent;
        }

        //    long end = System.nanoTime() - middle;
        //    start = middle - start;
        //    System.out.println("Find nodes: " + (start / 1000) + "us. Calc values: " + (end / 1000) + "us. Hops: " + hops + " ->");

        return new QueryResult(max, sum);
    }
}

/**
 * A node that is part of a priority-heap and an independent AVL Tree (for hashing) at the same time.
 * Extends TrigramAndCount.
 *
 * Created by Daniel Pujiula Buhl on 05.07.2015.
 */
class AVLPriorityHeapNode extends TrigramAndCount {

    // TrigramAndCount

    AVLPriorityHeapNode(String w1, String w2, String w3, int count) {
        super(w1, w2, w3, count);
    }
/*
    DoubleBinMaxHeapTreeNode(TrigramAndCount tc) {
        super(tc.getW1(), tc.getW2(), tc.getW3(), tc.getCount());
    }
*/
    // TreeNode

    AVLPriorityHeapNode parent = null, left = null, right = null;
    AVLPriorityHeapNode hash_parent = null, hash_left = null, hash_right = null;
    int id = -1;
    /* Accumulated Count */
    long cumCount = 0;

    private TreeBalance balance = TreeBalance.BALANCED;

    public TreeBalance getBalance() {
        return balance;
    }

    public void setBalance(TreeBalance balance) {
        //System.out.println("New Balance of " + key + " is " + balance);
        this.balance = balance;
    }
}

/**
 * A tree with methods that enable it to be handled as a Priority Heap as well as a AVL Tree for a hashMap.
 *
 * Created by Daniel Pujiula Buhl on 05.07.2015.
 */
class AVLPriorityHeap {

    AVLPriorityHeapNode root = null;

    public void bstInsert(AVLPriorityHeapNode node) {
        // Wir brauchen einen Schluessel um zu sortieren
        if(node == null) throw new NullPointerException("Key must not be null.");
        // Suche einen freien Platz
        AVLPriorityHeapNode root = this.root, parent = null;

        boolean smaller = true; /* will be overridden before use */

        // "smaller" needs to be initialised for the compiler not to complain.
        // in fact, in the case when smaller is used outside of the while-loop,
        // that is only if parent != null. But in that case, the while loop has been
        // executed at least once and thus smaller has been initialised. So the
        // following value has no significance whatsoever.

        while(root != null) {
            parent = root;
            smaller = node.compareTo(root) < 0;
            if(smaller) {
                root = root.left;
            } else {
                root = root.right;
            }
        }
        // Einfuegen
        node.parent = parent;
        if(parent == null) { // Baum war leer; neue Wurzel
            this.root = node;
        }
        else if (smaller) {
            parent.left = node;
        }
        else {
            parent.right = node;
        }
    }

    public AVLPriorityHeapNode bstSearch(Trigram key) {
        // Wir brauchen einen Schluessel um zu sortieren
        if(key == null) throw new NullPointerException("Key must not be null.");
        // durchsuche den Baum nach dem Schluessel
        AVLPriorityHeapNode current = root;
        int c;
        while(current != null) {
            c = current.compareTo(key);
            if(c < 0)
                current = current.left;
            else if(c > 0)
                current = current.right;
            else return current;
        }
        // nicht gefunden
        return null;
    }

    public AVLPriorityHeapNode hashSearch(Trigram key) {
        // Wir brauchen einen Schluessel um zu sortieren
        if(key == null) throw new NullPointerException("Key must not be null.");
        // durchsuche den Baum nach dem Schluessel
        AVLPriorityHeapNode current = root;
        int c;
        while(current != null) {
            c = current.compareTo(key);
            if(c < 0)
                current = current.hash_left;
            else if(c > 0)
                current = current.hash_right;
            else return current;
        }
        // nicht gefunden
        return null;
    }

    /* if node does not exist in tree, return last found node instead of null
    * before=true -> search nearest neighbour that is <= key,
    * before=false -> search nearest neighbor that ist >= key */
    public AVLPriorityHeapNode bstSearchNearestNeighbour(Trigram key, boolean before) {
        // Wir brauchen einen Schluessel um zu sortieren
        if(key == null) throw new NullPointerException("Key must not be null.");

        //System.out.println("--- Search Nearest Successor of " + key.toString() + " ---");

        // durchsuche den Baum nach dem Schluessel
        AVLPriorityHeapNode current = root, last = null;
        int c;
        while(current != null) {
            //System.out.print("Current:" + current + " is ");
            c = key.compareTo(current);
            //System.out.println((c<0) ? "smaller" : (c==0) ? " greater" : "equal" );
            if(c < 0) {
                if (!before) last = current;
                current = current.left;
            }
            else if(c > 0) {
                if (before) last = current;
                current = current.right;
            }
            else return current;
        }
        // nicht gefunden
        return last;
    }

    public AVLPriorityHeapNode bstMin(AVLPriorityHeapNode node) {
        if(node == null) throw new NullPointerException("No minimum of empty tree!");
        while(node.left != null) {
            node = node.left;
        }
        return node;
    }

    public AVLPriorityHeapNode bstMax(AVLPriorityHeapNode node) {
        if(node == null) throw new NullPointerException("No minimum of empty tree!");
        while(node.right != null) {
            node = node.right;
        }
        return node;
    }

    public AVLPriorityHeapNode bstSucc(AVLPriorityHeapNode node) {
        if(node.right != null) {
            return bstMin(node.right);
        }

        while (node.parent != null && node.parent.right == node) {
            node = node.parent;
        }

        return node.parent;
    }

    public void bstReplace(AVLPriorityHeapNode old, AVLPriorityHeapNode node) {
        if (node != null) {
            node.parent = old.parent;
        }
        if (old.parent == null) // war die Wurzel
        {
            this.root = node;
        }
        else if (old == old.parent.left) {
            //war linkes Kind
            old.parent.left = node;
        }
        else {
            // rechtes Kind
            old.parent.right = node;
        }
    }

    public void bstSwap(AVLPriorityHeapNode old, AVLPriorityHeapNode node) {
        //linken Teilbaum uebernehmen
        node.left = old.left;
        if(node.left != null) {
            node.left.parent = node;
        }
        // rechten Teilbaum
        if(node.right != null) {
            node.right.parent = node;
        }
        // Knoten einfuegen
        bstReplace(old, node);
    }
    /*
        public void bstDel(DoubleBinMaxHeapTreeNode node) {
            if (node.left != null && node.right != null) {
                DoubleBinMaxHeapTreeNode tmp = bstMin(node.right);
                bstDel(tmp); // hoechstens ein Kind, rechts
                bstSwap(node, tmp);
            } else if (node.left != null) { // ein Kind, links
                bstReplace(node, node.left);
            } else { // ein Kind oder kein Kind (node.right == null
                bstReplace(node, node.right);
            }
        }
    */
    void leftRotate(AVLPriorityHeapNode node1) {
        AVLPriorityHeapNode node2 = node1.right;
        //inneren Baum verschieben
        node1.right = node2.left;
        if (node1.right != null) {
            node1.right.parent = node1;
        }
        node2.parent = node1.parent;
        if(node1.parent == null) {
            root = node2;
        }
        else if (node1 == node1.parent.left) {
            node2.parent.left = node2;
        }
        else {
            node2.parent.right = node2;
        }
        node2.left = node1;
        node1.parent = node2;
    }

    void rightRotate(AVLPriorityHeapNode node1) {
        AVLPriorityHeapNode node2 = node1.left;
        //inneren Baum verschieben
        node1.left = node2.right;
        if (node1.left != null) {
            node1.left.parent = node1;
        }
        node2.parent = node1.parent;
        if(node1.parent == null) {
            root = node2;
        }
        else if (node1 == node1.parent.right) {
            node2.parent.right = node2;
        }
        else {
            node2.parent.left = node2;
        }
        node2.right = node1;
        node1.parent = node2;
    }

    void hashLeftRotate(AVLPriorityHeapNode node1) {
        AVLPriorityHeapNode node2 = node1.hash_right;
        //inneren Baum verschieben
        node1.hash_right = node2.hash_left;
        if (node1.hash_right != null) {
            node1.hash_right.hash_parent = node1;
        }
        node2.hash_parent = node1.hash_parent;
        if(node1.hash_parent == null) {
            root = node2;
        }
        else if (node1 == node1.hash_parent.hash_left) {
            node2.hash_parent.hash_left = node2;
        }
        else {
            node2.hash_parent.hash_right = node2;
        }
        node2.hash_left = node1;
        node1.hash_parent = node2;
    }

    void hashRightRotate(AVLPriorityHeapNode node1) {
        AVLPriorityHeapNode node2 = node1.hash_left;
        //inneren Baum verschieben
        node1.hash_left = node2.hash_right;
        if (node1.hash_left != null) {
            node1.hash_left.hash_parent = node1;
        }
        node2.hash_parent = node1.hash_parent;
        if(node1.hash_parent == null) {
            root = node2;
        }
        else if (node1 == node1.hash_parent.hash_right) {
            node2.hash_parent.hash_right = node2;
        }
        else {
            node2.hash_parent.hash_left = node2;
        }
        node2.hash_right = node1;
        node1.hash_parent = node2;
    }

    public void hashInsert(AVLPriorityHeapNode node) {
        /* bstInsert(node); */

        // es ergibt keinen Sinn, eine leere node einzufuegen;
        if(node == null) return;

        // Suche einen freien Platz
        AVLPriorityHeapNode root = this.root, parent = null;

        boolean smaller = true; /* will be overridden before use */

        // "smaller" needs to be initialised for the compiler not to complain.
        // in fact, in the case when smaller is used outside of the while-loop,
        // that is only if parent != null. But in that case, the while loop has been
        // executed at least once and thus smaller has been initialised. So the
        // following value has no significance whatsoever.

        // Now to simplify AVL-Algorithms:
        // remember last unbalanced node(s) we stumbled upon so we can directly determine
        // deepestUnbalancedNode :)

        while (root != null) {
            parent = root;
            smaller = node.compareTo(root) < 0;
            if (smaller) {
                root = root.hash_left;
            } else  {
                root = root.hash_right;
            } // ACHTUNG: Prueft nicht ob key bereits existiert
        }
        // Einfuegen
        node.hash_parent = parent;
        if(parent == null) { // Baum war leer; neue Wurzel
            this.root = node;
            return;
        }
        else if (smaller) {
            parent.hash_left = node;
        }
        else {
            parent.hash_right = node;
        }

        //System.out.println("Inserted " + node.key + " as child of " + parent.key);

        do {
            if(node.hash_parent.getBalance() == TreeBalance.BALANCED) {
                //links eingefuegt
                if(node.hash_parent.hash_left == node) {
                    node.hash_parent.setBalance(TreeBalance.LEFT);
                }
                //rechts eingefuegt
                else {
                    node.hash_parent.setBalance(TreeBalance.RIGHT);
                }
            }
            // bereits linkes Uebergewicht
            else if(node.hash_parent.getBalance() == TreeBalance.LEFT) {
                // und links eingefuegt
                if(node.hash_parent.hash_left == node) {
                    // Parent ist unbalanciert
                    // ACTION = RIGHTROTATE
                    hashBalanceLeft(node.hash_parent);
                    return;
                }
                else {
                    // parent wurde balanciert
                    node.hash_parent.setBalance(TreeBalance.BALANCED);
                    // ACTION = NOROTATE
                    return;
                }
            }
            // parent's balance is RIGHT
            else {
                if(node.hash_parent.hash_right == node) {
                    // ACTION = LEFTROTATE
                    hashBalanceRight(node.hash_parent);
                    return;
                }
                else {
                    node.hash_parent.setBalance(TreeBalance.BALANCED);
                    return;
                }
            }
            node = node.hash_parent;
        } while (node.hash_parent != null);
    }

    /* AVL Balance left-of-left and right-of-left case */
    private void hashBalanceLeft(AVLPriorityHeapNode node) {
        //System.out.println(this);
        if(node.hash_left.getBalance() == TreeBalance.LEFT) {
            hashRightRotate(node);
            node.setBalance(TreeBalance.BALANCED);
            node.hash_parent.setBalance(TreeBalance.BALANCED);
        }
        else {
            if(node.hash_left.hash_right.getBalance() == TreeBalance.BALANCED) {
                node.setBalance(TreeBalance.BALANCED);
                node.hash_left.setBalance(TreeBalance.BALANCED);
                node.hash_left.hash_right.setBalance(TreeBalance.BALANCED);
            }
            else if(node.hash_left.hash_right.getBalance() == TreeBalance.RIGHT) {
                node.setBalance(TreeBalance.BALANCED); // A
                node.hash_left.setBalance(TreeBalance.LEFT); // B
                node.hash_left.hash_right.setBalance(TreeBalance.BALANCED); // C
            } else {
                node.setBalance(TreeBalance.RIGHT); // A
                node.hash_left.setBalance(TreeBalance.BALANCED); // B
                node.hash_left.hash_right.setBalance(TreeBalance.BALANCED); // C
            }
            hashLeftRotate(node.hash_left);
            hashRightRotate(node);
        }
    }

    /* AVL balance right-of-right and left-of-right case */
    private void hashBalanceRight(AVLPriorityHeapNode node) {
        //System.out.println(this);
        if(node.hash_right.getBalance() == TreeBalance.RIGHT) {
            hashLeftRotate(node);
            node.setBalance(TreeBalance.BALANCED);
            node.hash_parent.setBalance(TreeBalance.BALANCED);
        }
        else {
            if(node.hash_right.hash_left.getBalance() == TreeBalance.BALANCED) {
                node.setBalance(TreeBalance.BALANCED);
                node.hash_right.setBalance(TreeBalance.BALANCED);
                node.hash_right.hash_left.setBalance(TreeBalance.BALANCED);
            }
            else if(node.hash_right.hash_left.getBalance() == TreeBalance.LEFT) {
                node.setBalance(TreeBalance.BALANCED); // A
                node.hash_right.setBalance(TreeBalance.RIGHT); // B
                node.hash_right.hash_left.setBalance(TreeBalance.BALANCED); // C
            } else {
                node.setBalance(TreeBalance.LEFT); // A
                node.hash_right.setBalance(TreeBalance.BALANCED); // B
                node.hash_right.hash_left.setBalance(TreeBalance.BALANCED); // C
            }
            hashRightRotate(node.hash_right);
            hashLeftRotate(node);
        }
    }

    public void priorityHeapInsert(AVLPriorityHeapNode node) {
        /* bstInsert(node); */

        // es ergibt keinen Sinn, eine leere node einzufuegen;
        if(node == null) return;

        // Suche einen freien Platz
        AVLPriorityHeapNode root = this.root, parent = null;

        int c  = 0; /* will be overridden before use */

        // "smaller" needs to be initialised for the compiler not to complain.
        // in fact, in the case when smaller is used outside of the while-loop,
        // that is only if parent != null. But in that case, the while loop has been
        // executed at least once and thus smaller has been initialised. So the
        // following value has no significance whatsoever.

        // Now to simplify AVL-Algorithms:
        // remember last unbalanced node(s) we stumbled upon so we can directly determine
        // deepestUnbalancedNode :)

        while (root != null) {
            parent = root;
            c = node.compareTo(root);
            if (c < 0) {
                root = root.left;
            } else if (c > 0) {
                root = root.right;
            } else {
                return; //Existiert bereits
            }
        }
        // Einfuegen
        node.parent = parent;
        if(parent == null) { // Baum war leer; neue Wurzel
            this.root = node;
            return;
        }
        else if (c < 0) {
            parent.left = node;
        }
        else {
            parent.right = node;
        }

        //System.out.println("Inserted " + node.key + " as child of " + parent.key);

        // recreate MaxHeap

        while (node.parent != null && node.getCount() > node.parent.getCount()) {
            if(node.parent.left == node) {
                rightRotate(node.parent);
            }
            else {
                leftRotate(node.parent);
            }
        }
    }

    public int hash_height() {
        return hash_height(root);
    }

    public int hash_height(AVLPriorityHeapNode node) {
        if(node==null) return 0;
        return 1 + Math.max(hash_height(node.hash_left), hash_height(node.hash_right));
    }
}

enum TreeBalance {
    BALANCED, LEFT, RIGHT
}
