package i2;

import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Random;

import trigrams.QueryResult;
import trigrams.Trigram;
import trigrams.TrigramAndCount;

public class BinarySearchTrigramStorage extends SimpleTrigramStorage {
	
	@Override
	public void build() {
		java.util.Collections.sort(this.trigrams);
	}
	
	@Override
	public QueryResult query(Trigram trigram1, Trigram trigram2) {
		long max = 0;
		long sum = 0;
		
		// Find the two points.
		int first = this.findPositionBinarySearch(trigram1, true);
		int second = this.findPositionBinarySearch(trigram2, false);

		for (int i = first; i < second; i++) {
			TrigramAndCount trigram = this.trigrams.get(i);
			max = Math.max(max, trigram.getCount());
			sum += trigram.getCount();
		}
		
		return new QueryResult(max, sum);
	}
	
	private int findPositionBinarySearch(Trigram trigram, boolean lowerBound) {
		int left = 0;
		int right = this.trigrams.size();
		
		while (left != right) {
			int mid = left + (right - left) / 2;
			if ((lowerBound && this.trigrams.get(mid).compareTo(trigram) < 0) || (!lowerBound && this.trigrams.get(mid).compareTo(trigram) <= 0)) {
				left = mid + 1;
			} else {
				right = mid;
			}
		}
		
		return left;
	}
	
	public static int MAX_WORD_LENGTH = 10;
	
	private String generateRandomWord() {
		StringBuilder builder = new StringBuilder();
		
		Random rand = new Random();
		
		int length = rand.nextInt(MAX_WORD_LENGTH) + 1;
		
		for (int index = 0; index < length; ++index) {
			int charIndex = rand.nextInt(36);
			
			if (charIndex < 10) {
				charIndex += 48;
			} else {
				charIndex += 87;
			}
			
			builder.append((char)charIndex);
		}
		
		return builder.toString();	
	}
	
	private Trigram generateRandomTrigram() {
		return new Trigram(generateRandomWord(), generateRandomWord(), generateRandomWord());
	}
	
	enum QuerySize { VERY_LARGE, LARGE, MEDIUM, SMALL }
	enum QueryVocabulary { KNOWN, UNKNOWN }

	public void generateRandomQueryData(Writer writer, QuerySize size, QueryVocabulary vocab) throws IOException {
		if (vocab == QueryVocabulary.KNOWN) {
			int first = 0;
			int second = 0;
			Random rand = new Random();
			if (size == QuerySize.VERY_LARGE) {
				first = rand.nextInt(this.trigrams.size() - 1);
				second = rand.nextInt(this.trigrams.size() / 2 + 1) + first;
			} if (size == QuerySize.LARGE) {
				first = rand.nextInt(this.trigrams.size() - 1);
				second = rand.nextInt(this.trigrams.size() / 10 + 1) + first;
			} else if (size == QuerySize.MEDIUM) {
				first = rand.nextInt(this.trigrams.size() - 1);
				second = rand.nextInt(this.trigrams.size() / 100 + 1) + first; 
			} else if (size == QuerySize.SMALL) {
				first = rand.nextInt(this.trigrams.size() - 1);
				second = rand.nextInt(this.trigrams.size() / 1000 + 1) + first; 
			}
			second = Math.min(second, this.trigrams.size() - 1);			
			generateQueryData(writer, first, second);
		} else {
			Trigram trigram1 = null;
			Trigram trigram2 = null;
			do {
				trigram1 = generateRandomTrigram();
				trigram2 = generateRandomTrigram();
			} while (trigram1.compareTo(trigram2) >= 0);
			
			QueryResult result = query(trigram1, trigram2);
			writer.write(trigram1.getW1() + " " + trigram1.getW2() + " " + trigram1.getW3() + " " + trigram2.getW1() + " " + trigram2.getW2() + " " + trigram2.getW3() + " " + result.getMax() + " " + result.getSum() + "\n");
		}
	}
	
	public void generateQueryData(Writer writer, int index1, int index2) throws IOException {
		Trigram trigram1 = this.trigrams.get(index1);
		Trigram trigram2 = this.trigrams.get(index2);
		QueryResult result = query(trigram1, trigram2);
		writer.write(trigram1.getW1() + " " + trigram1.getW2() + " " + trigram1.getW3() + " " + trigram2.getW1() + " " + trigram2.getW2() + " " + trigram2.getW3() + " " + result.getMax() + " " + result.getSum() + "\n");
	}
	
	public static int NUMBER_OF_WARMUP_QUERIES = 00;
	public static int NUMBER_OF_SMALL_QUERIES = 20000;
	public static int NUMBER_OF_MEDIUM_QUERIES = 20000;
	public static int NUMBER_OF_LARGE_QUERIES = 20000;
	public static int NUMBER_OF_VERY_LARGE_QUERIES = 20000;
	public static int NUMBER_OF_UNKNOWN_VOCAB_QUERIES = 20000;
	
	public void generateQueries(String queriesFilename) throws IOException {
		Writer writer = null;
		try {
			writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(queriesFilename), "utf-8"));
			
			// Create warmup queries.
			writer.write("[warmup]\n");
			for (int index = 0; index < NUMBER_OF_WARMUP_QUERIES / 3; ++index) {
				generateRandomQueryData(writer, QuerySize.SMALL, QueryVocabulary.KNOWN);
			}
			for (int index = 0; index < NUMBER_OF_WARMUP_QUERIES / 3; ++index) {
				generateRandomQueryData(writer, QuerySize.MEDIUM, QueryVocabulary.KNOWN);
			}
			for (int index = 0; index < NUMBER_OF_WARMUP_QUERIES / 3; ++index) {
				generateRandomQueryData(writer, QuerySize.LARGE, QueryVocabulary.KNOWN);
			}

			// Generate small queries.
			writer.write("[small]\n");
			for (int index = 0; index < NUMBER_OF_SMALL_QUERIES; ++index) {
				generateRandomQueryData(writer, QuerySize.SMALL, QueryVocabulary.KNOWN);
			}
			
			// Generate medium-sized queries.
			writer.write("[medium]\n");
			for (int index = 0; index < NUMBER_OF_MEDIUM_QUERIES; ++index) {
				generateRandomQueryData(writer, QuerySize.MEDIUM, QueryVocabulary.KNOWN);
			}
			
			// Generate large queries.
			writer.write("[large]\n");
			for (int index = 0; index < NUMBER_OF_LARGE_QUERIES; ++index) {
				generateRandomQueryData(writer, QuerySize.LARGE, QueryVocabulary.KNOWN);
			}
			
			// Generate very large queries. Start with the full query.
			writer.write("[verylarge]\n");
			generateQueryData(writer, 0, this.trigrams.size() - 1);
			for (int index = 0; index < NUMBER_OF_VERY_LARGE_QUERIES; ++index) {
				generateRandomQueryData(writer, QuerySize.VERY_LARGE, QueryVocabulary.KNOWN);
			}
			generateQueryData(writer, 0, this.trigrams.size() - 1);
			
			writer.write("[unknown]\n");
			for (int index = 0; index < NUMBER_OF_UNKNOWN_VOCAB_QUERIES; ++index) {
				generateRandomQueryData(writer, QuerySize.SMALL, QueryVocabulary.UNKNOWN);
			}

		} catch (FileNotFoundException e) {
			// This cannot happen.
		} finally {
			writer.close();
		}
	}
	
}
