package i2;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import trigrams.IntegerTrigram;
import trigrams.IntegerTrigramAndCount;
import trigrams.QueryResult;
import trigrams.Trigram;
import trigrams.TrigramAndCount;
import trigrams.TrigramStorage;


public class OptimizedTrigramStorage implements TrigramStorage {

	public OptimizedTrigramStorage() {
		trigrams = new ArrayList<TrigramAndCount>();
		words = new TreeMap<String, String>();
	}
	
	@Override
	public void insert(TrigramAndCount trigram) {
		String w1 = words.get(trigram.getW1());
		if (w1 == null) {
			words.put(trigram.getW1(), trigram.getW1());
			w1 = trigram.getW1();
		}
		String w2 = words.get(trigram.getW2());
		if (w2 == null) {
			w2 = words.put(trigram.getW2(), trigram.getW2());
			w2 = trigram.getW2();
		}
		String w3 = words.get(trigram.getW3());
		if (w3 == null) {
			w3 = words.put(trigram.getW3(), trigram.getW3());
			w3 = trigram.getW3();
		}
		trigrams.add(new TrigramAndCount(w1, w2, w3, trigram.getCount()));
	}

	@Override
	public void build() {
		// We start by translating the string-based trigrams to integer-based ones.
		
		// To do so, we first build a hash table that maps the elements of the tree
		// set to their 'index' in the ordered set.
		dictionary = new HashMap<String, Long>();
		reverseDictionary = new HashMap<Long, String>();
		int intIndex = 0;
		wordarray = new String[words.size()];
		for (Map.Entry<String, String> word : words.entrySet()) {
			dictionary.put(word.getValue(), (long)intIndex);
			reverseDictionary.put((long)intIndex, word.getValue());
			wordarray[intIndex] = word.getValue();
			++intIndex;
		}
		
		// Set the word set to null to make it disposable.
		words = null;
		
		// Now we can build an array with the integer-based trigrams.
		itrigrams = new IntegerTrigramAndCount[trigrams.size()]; 
		
		intIndex = 0;
		for (TrigramAndCount trigram : trigrams) {
			itrigrams[intIndex] = new IntegerTrigramAndCount(
					dictionary.get(trigram.getW1()),
					dictionary.get(trigram.getW2()),
					dictionary.get(trigram.getW3()),
					trigram.getCount()
					);
			++intIndex;
		}
		
		// Set the original trigrams to null to make them disposable.
		trigrams = null;
		
		// Then we can sort the array of integer-based trigrams.
		OptimizedTrigramStorage.quickSort(itrigrams);
				
		// Next, we build a variation of a skip-list. We map every element to the index
		// that possesses the next largest count.
		maxSkipIndices = new int[itrigrams.length];
		maxSkipIndices[itrigrams.length - 1] = itrigrams.length;
		int intIndex2 = 0;
		for (intIndex = itrigrams.length - 1; intIndex >= 0; --intIndex) {
			long currentCount = itrigrams[intIndex].getCount();
						
			// Find the next element with a higher count.
			for (intIndex2 = intIndex + 1; intIndex2 < itrigrams.length; ) {
				// Check the position intIndex2 for a higher value.
				if (itrigrams[intIndex2].getCount() > currentCount) {
					maxSkipIndices[intIndex] = intIndex2;
					break;
				}
				
				// If the value was not higher (yet), we follow the skip list pointer.
				intIndex2 = maxSkipIndices[intIndex2];
			}
			
			// If there was no element with higher count, set the value "manually".
			if (intIndex2 == itrigrams.length) {
				maxSkipIndices[intIndex] = intIndex2;
			}
		}
		
		// Finally, we fill the running sum array.
		runningSum = new long[itrigrams.length + 1];
		intIndex = 0;
		long sum = 0;
		for (IntegerTrigramAndCount itrigram : itrigrams) {
			runningSum[intIndex] = sum;
			sum += itrigram.getCount();
			++intIndex;
		}
		runningSum[itrigrams.length] = sum;
	}

	private int findPositionBinarySearch(IntegerTrigram itrigram, boolean lowerBound) {
		int left = 0;
		int right = this.itrigrams.length;
		
		while (left != right) {
			int mid = left + (right - left) / 2;
			if ((lowerBound && this.itrigrams[mid].compareTo(itrigram) < 0) || (!lowerBound && this.itrigrams[mid].compareTo(itrigram) <= 0)) {
				left = mid + 1;
			} else {
				right = mid;
			}
		}
		
		return left;
	}
	
	private int findPositionBinarySearch(String word, boolean lowerBound) {
		int left = 0;
		int right = this.wordarray.length;
		
		while (left != right) {
			int mid = left + (right - left) / 2;
			if ((lowerBound && this.wordarray[mid].compareTo(word) < 0) || (!lowerBound && this.wordarray[mid].compareTo(word) <= 0)) {
				left = mid + 1;
			} else {
				right = mid;
			}
		}
		
		return left;
	}
	
	private IntegerTrigram convertToIntegerTrigram(Trigram trigram, boolean lowerBound) {
		if (lowerBound) {
			int w1 = 0;
			Long w1Lookup = dictionary.get(trigram.getW1());
			boolean w1Found = false;
			if (w1Lookup != null) {
				w1 = (int)(long)w1Lookup;
				w1Found = true;
			} else {
				w1 = this.findPositionBinarySearch(trigram.getW1(), lowerBound);
			}
			
			int w2 = -1;
			boolean w2Found = false;
			if (w1Found) {
				Long w2Lookup = dictionary.get(trigram.getW2());
				if (w2Lookup != null) {
					w2 = (int)(long)w2Lookup;
					w2Found = true;
				} else {
					w2 = this.findPositionBinarySearch(trigram.getW2(), lowerBound);
				}
			}
			
			int w3 = -1;
			if (w2Found) {
				Long w3Lookup = dictionary.get(trigram.getW3());
				if (w3Lookup != null) {
					w3 = (int)(long)w3Lookup;
				} else {
					w3 = this.findPositionBinarySearch(trigram.getW3(), lowerBound);
				}
			}
			
			return new IntegerTrigram(w1, w2, w3);
		} else {
			int w1 = 0;
			Long w1Lookup = dictionary.get(trigram.getW1());
			boolean w1Found = false;
			if (w1Lookup != null) {
				w1 = (int)(long)w1Lookup;
				w1Found = true;
			} else {
				w1 = this.findPositionBinarySearch(trigram.getW1(), lowerBound) - 1;
			}
			
			int w2 = this.wordarray.length;
			boolean w2Found = false;
			if (w1Found) {
				Long w2Lookup = dictionary.get(trigram.getW2());
				if (w2Lookup != null) {
					w2 = (int)(long)w2Lookup;
					w2Found = true;
				} else {
					w2 = this.findPositionBinarySearch(trigram.getW2(), lowerBound) - 1;
				}
			}
			
			int w3 = this.wordarray.length;
			if (w2Found) {
				Long w3Lookup = dictionary.get(trigram.getW3());
				if (w3Lookup != null) {
					w3 = (int)(long)w3Lookup;
				} else {
					w3 = this.findPositionBinarySearch(trigram.getW3(), lowerBound) - 1;
				}
			}
			
			return new IntegerTrigram(w1, w2, w3);
		}
	}
	
	@Override
	public QueryResult query(Trigram trigram1, Trigram trigram2) {
		int maxIndex = 0, nextMaxIndex = 0;
		long max = 0;
		long sum = 0;
		
		IntegerTrigram itrigram1 = this.convertToIntegerTrigram(trigram1, true);
		IntegerTrigram itrigram2 = this.convertToIntegerTrigram(trigram2, false);
		
		// Find the two points.
		int first = nextMaxIndex = maxIndex = this.findPositionBinarySearch(itrigram1, true);
		int second = this.findPositionBinarySearch(itrigram2, false);

		if (first < second) {
			// Use the skip-list to determine the maximum.
			while (nextMaxIndex < second) {
				maxIndex = nextMaxIndex; 
				nextMaxIndex = maxSkipIndices[nextMaxIndex];
			}
			max = itrigrams[maxIndex].getCount();
			
			// Compute the sum via the running sum.
			sum = runningSum[second] - runningSum[first];
		}
		
		return new QueryResult(max, sum);
	}

	private static void swap(IntegerTrigramAndCount[] itrigrams, int i, int j) {
		IntegerTrigramAndCount tmp = itrigrams[i];
		itrigrams[i] = itrigrams[j];
		itrigrams[j] = tmp;
	}
	
	private static void quickSort(IntegerTrigramAndCount[] itrigrams, int left, int right) {
		if (left < right) {
			// Take a pivot element roughly in the middle.
	        int pivotIndex = (left + right) / 2;
	        IntegerTrigramAndCount pivotValue = itrigrams[pivotIndex];

			int i = left;
			int j = right;
	    	
	    	while (i <= j) {
		    	while (itrigrams[i].compareTo(pivotValue) < 0) {
		    		i++;
		    	}
		    	while (itrigrams[j].compareTo(pivotValue) > 0) {
		    		j--;
		        }
		    	if (i <= j) {
		    		swap(itrigrams, i, j);
		    		i++;
		    		j--;
		    	}
		    }
	    	if (left < j) {
	    		quickSort(itrigrams, left, j);
			}
	    	if (i < right) {
	    		quickSort(itrigrams, i, right);
	    	}
		}
	}
	
	private static void quickSort(IntegerTrigramAndCount[] itrigrams) {
		quickSort(itrigrams, 0, itrigrams.length - 1);
	}
	
	// This is the structure into which we insert the data in the insertion phase.
	private ArrayList<TrigramAndCount> trigrams;
	
	// We also keep a set of the words, so we can map them to integers later.
	private TreeMap<String, String> words;
	
	// These are the structures that are built up by the <code>build()</code> method.
	// They are used to answer the actual queries later.
	
	// A mapping from the words to their lexicographical position.
	HashMap<String, Long> dictionary;
	HashMap<Long, String> reverseDictionary;
	
	// Also store a (lexicographically sorted) array of words.
	String wordarray[];
	
	// An array that stores the running sum.
	long runningSum[];
	
	// An array that stores for each position the next index with a *higher* count.
	int maxSkipIndices[];
	
	// An array storing the actual trigrams (integer-based) and their counts.
	IntegerTrigramAndCount[] itrigrams;
	
}
